/**
 * Ext GWT 3.0.0-beta4 - Ext for GWT
 * Copyright(c) 2007-2011, Sencha, Inc.
 * licensing@sencha.com
 *
 * http://sencha.com/license
 */
package com.sencha.gxt.data.client.loader;

import com.google.gwt.core.client.Callback;
import com.google.gwt.storage.client.Storage;
import com.sencha.gxt.data.client.loader.StorageWriteProxy.KeyAndValue;
import com.sencha.gxt.data.shared.loader.DataProxy;

/**
 * Proxy to push key/value string pairs into local html5 browser storage. Both
 * key and value must be populated prior to
 * 
 */
public class StorageWriteProxy implements DataProxy<KeyAndValue, Void> {

  /**
   * Defines a key / value pair.
   */
  public static class KeyAndValue {
    String key;
    String value;
  }

  private final Storage storage;

  /**
   * Creates a storage write proxy that saves a key and value to HTML5 browser
   * based storage.
   * 
   * @param session true to use session storage, false to use local storage
   */
  public StorageWriteProxy(boolean session) {
    this(session ? Storage.getSessionStorageIfSupported() : Storage.getLocalStorageIfSupported());
  }

  /**
   * Creates a storage write proxy that saves a key and value to the given HTML5
   * browser based storage.
   * 
   * @param storage the browser based storage
   */
  public StorageWriteProxy(Storage storage) {
    assert storage != null : "Storage may not be null";
    this.storage = storage;
  }

  @Override
  public void load(KeyAndValue data, Callback<Void, Throwable> callback) {
    storage.setItem(data.key, data.value);
    callback.onSuccess(null);
  }

}
