/**
 * Ext GWT 3.0.0-beta4 - Ext for GWT
 * Copyright(c) 2007-2011, Sencha, Inc.
 * licensing@sencha.com
 *
 * http://sencha.com/license
 */
package com.sencha.gxt.data.shared.loader;

import com.google.gwt.core.client.Callback;
import com.google.web.bindery.requestfactory.shared.Receiver;
import com.google.web.bindery.requestfactory.shared.ServerFailure;

/**
 * An abstract request factory proxy that supports processing results using
 * either a RPC style <code>Callback</code> or a <code>RequestFactory</code>
 * <code>Receiver</code>.
 * 
 * @param <C> the type of data used to configure the load from the proxy
 * @param <D> the type of data being returned by the data proxy
 */
public abstract class RequestFactoryProxy<C, D> implements DataProxy<C, D> {
  public final void load(C loadConfig, final Callback<D, Throwable> callback) {
    load(loadConfig, new Receiver<D>() {
      @Override
      public void onFailure(ServerFailure error) {
        callback.onFailure(new RuntimeException(error.getMessage()));
      }

      public void onSuccess(D response) {
        callback.onSuccess(response);
      }
    });
  }

  /**
   * Process results using a <code>Receiver</code>.
   * 
   * @param loadConfig the load config object to be passed to server
   * @param receiver the data callback
   */
  public abstract void load(C loadConfig, Receiver<? super D> receiver);
}
